@@ -38,12 +38,12 @@ HB_BEGIN_DECLS
 
 #define HB_VERSION_MAJOR 0
 #define HB_VERSION_MINOR 9
-#define HB_VERSION_MICRO 29
+#define HB_VERSION_MICRO 30
 
-#define HB_VERSION_STRING "0.9.29"
+#define HB_VERSION_STRING "0.9.30"
 
-#define HB_VERSION_CHECK(major,minor,micro) \
-	((major)*10000+(minor)*100+(micro) >= \
+#define HB_VERSION_ATLEAST(major,minor,micro) \
+	((major)*10000+(minor)*100+(micro) <= \
 	 HB_VERSION_MAJOR*10000+HB_VERSION_MINOR*100+HB_VERSION_MICRO)
 
 
@@ -56,9 +56,9 @@ const char *
 hb_version_string (void);
 
 hb_bool_t
-hb_version_check (unsigned int major,
-		  unsigned int minor,
-		  unsigned int micro);
+hb_version_atleast (unsigned int major,
+		    unsigned int minor,
+		    unsigned int micro);
 
 
 HB_END_DECLS
